@@ -8,7 +8,9 @@ use HTML::Template::Compiled::Expression qw(:expressions);
 use HTML::Template::Compiled::Utils qw(:walkpath);
 use File::Basename qw(dirname);
 
-our $VERSION = '0.18';
+our $VERSION = '0.20';
+
+our $DISABLE_NEW_ALIAS = 0;
 
 use Carp qw(croak carp);
 use constant D             => 0;
@@ -31,6 +33,7 @@ use constant T_INCLUDE_VAR => 'INCLUDE_VAR';
 use constant T_INCLUDE_STRING => 'INCLUDE_STRING';
 use constant T_USE_VARS    => 'USE_VARS';
 use constant T_SET_VAR     => 'SET_VAR';
+use constant T_WRAPPER     => 'WRAPPER';
 
 use constant INDENT        => '    ';
 
@@ -54,34 +57,41 @@ sub get_tags       { $_[0]->[ATTR_TAGS] }
 sub set_name_re    { $_[0]->[ATTR_NAME_RE] = $_[1] }
 sub get_name_re    { $_[0]->[ATTR_NAME_RE] }
 
+our %ESCAPES;
+
 sub delete_subs {
     # delete all userdefined subs
-    my ($class) = @_;
-    no strict 'refs';
-    #warn __PACKAGE__." finding all subs\n";
-    for my $key (keys %{'HTML::Template::Compiled::Compiler::subs::'} ) {
-        my $sym = ${'HTML::Template::Compiled::Compiler::subs::'}{$key};
-        if (defined *{$sym}{CODE}) {
-            #warn __PACKAGE__." $key is code\n";
-            delete ${'HTML::Template::Compiled::Compiler::subs::'}{$key}
+    %ESCAPES = ();
+}
+
+sub setup_escapes {
+    my ($class, $plug_class, $escapes) = @_;
+    for my $key (keys %$escapes) {
+        my $def = $escapes->{$key};
+        my $sub;
+        if (ref $def eq 'HASH') {
+            $sub = $def->{code};
+            if (my $arguments = $def->{arguments} ) {
+                $ESCAPES{ $plug_class }->{ $key }->{arguments} = $arguments;
+            }
+        }
+        else {
+            $sub = $def;
+        }
+        if (ref $sub eq 'CODE') {
+            $ESCAPES{ $plug_class }->{ $key }->{code} = $sub;
+        }
+        else {
+            $ESCAPES{ $plug_class }->{ $key }->{code} = \&{ $sub };
         }
     }
 }
 
 sub add_escapes {
-    my ($self, $new_escapes) = @_;
+    my ($self, $plug_class, $new_escapes) = @_;
     my $escapes = $self->get_escapes;
-    for my $key (%$new_escapes) {
-        my $sub = $new_escapes->{$key};
-        if (ref $sub eq 'CODE') {
-            my $subname = "HTML::Template::Compiled::Compiler::subs::$key";
-            no strict 'refs';
-            *$subname = $sub;
-            $escapes->{$key} = $subname;
-        }
-        else {
-            $escapes->{$key} = $sub;
-        }
+    for my $key (keys %$new_escapes) {
+        $escapes->{ $key } = $plug_class;
     }
 }
 
@@ -127,12 +137,26 @@ sub _escape_expression {
         elsif ( $_ eq 'DUMP' ) {
             $exp = _expr_method( 'dump', _expr_literal('$t'), $exp, );
         }
-        elsif (my $sub = $escapes->{$_}) {
-            $exp = _expr_function( $sub, $exp );
+        elsif (my $plug_class = $escapes->{$_}) {
+            my $subref = "\$HTML::Template::Compiled::Compiler::ESCAPES\{'$plug_class'\}->\{'$_'\}->\{code\}";
+            my @args = $exp;
+            if (my $arguments = $ESCAPES{ $plug_class }->{ $_ }->{arguments}) {
+                @args = ();
+                for my $arg (@$arguments) {
+                    if ($arg eq 'var') {
+                        push @args, $exp;
+                    }
+                    elsif ($arg eq 'self') {
+                        push @args, "\$t->get_plugin('$plug_class')";
+                        #push @args, 23;
+                    }
+                }
+            }
+            $exp = HTML::Template::Compiled::Expression::SubrefCall->new( $subref, @args );
         }
-    } ## end for (@escapes)
+    }
     return ref $exp ? $exp->to_string : $exp;
-} ## end sub _escape_expression
+}
 
 sub init_name_re {
     my ($self, %args) = @_;
@@ -150,12 +174,15 @@ my %loop_context = (
     __first__   => '$__ix__ == $[',
     __last__    => '$__ix__ == $__size__',
     __odd__     => '!($__ix__ & 1)',
+    __even__    => '($__ix__ & 1)',
     __inner__   => '$__ix__ != $[ && $__ix__ != $__size__',
+    __outer__   => '$__ix__ == $[ || $__ix__ == $__size__',
     __key__     => '$__key__',
     __value__   => '$__value__',
     __break__   => '$__break__',
     __filename__ => '$t->get_file',
     __filenameshort__ => '$t->get_filename',
+    __wrapped__ => '$args->{wrapped}',
 );
 
 sub parse_var {
@@ -178,7 +205,6 @@ sub parse_var {
     }
 
 
-    # only allow '.', '/', '+', '-' and '_'
     if (!$t->validate_var($args{var})) {
         $t->get_parser->_error_wrong_tag_syntax(
             {
@@ -191,12 +217,13 @@ sub parse_var {
         );
     }
     if ( grep { defined $_ && $args{var} eq $_ } @$lexicals ) {
-        my $varstr = "\$lexi_$args{var}";
+        my $varstr = "\$HTML::Template::Compiled::_lexi_$args{var}";
         return $varstr;
     }
     my $lexi = join '|', grep defined, @$lexicals;
     my $varname = '$var';
     my $re = $self->get_name_re;
+#    warn __PACKAGE__.':'.__LINE__.": re: $re\n";
     #warn __PACKAGE__.':'.__LINE__.": ========== ($args{var})\n";
     my $root         = 0;
     my $up_stack = 0;
@@ -207,8 +234,12 @@ sub parse_var {
             return $lc;
         }
     }
-    if ($lexi and $args{var} =~ s/^($lexi)($re)/$2/) {
-        $initial_var = "\$lexi_$1";
+    # explicitly use aliases with '$' at the beginning
+    if (not $DISABLE_NEW_ALIAS and $args{var} =~ s/^\$(\w+)//) {
+        $initial_var = "\$HTML::Template::Compiled::_lexi_$1";
+    }
+    elsif ($lexi and $args{var} =~ s/^($lexi)($re)/$2/) {
+        $initial_var = "\$HTML::Template::Compiled::_lexi_$1";
     }
     elsif ( $args{var} =~ m/^_/ && $args{var} !~ m/^__(\w+)__$/ ) {
         $args{var} =~ s/^_//;
@@ -394,7 +425,7 @@ sub compile {
     }
     my $parser = $self->get_parser;
     my @p = $parser->parse($fname, $text);
-    if (my $df = $self->get_debug_file) {
+    if (my $df = $self->get_debug->{file}) {
         my $debugfile = $df =~ m/short/ ? $self->get_filename : $self->get_file;
         if ($df =~ m/start/) {
             unshift @p, 
@@ -415,23 +446,27 @@ sub compile {
     my $info = {}; # for query()
     my $info_stack = [$info];
 
+my $test = $self->get_debug->{options};
     # got this trick from perlmonks.org
     my $anon = D
-      || $self->get_debug ? qq{local *__ANON__ = "htc_$fname";\n} : '';
+      || ($self->get_debug->{options} & HTML::Template::Compiled::DEBUG_COMPILED()) ? qq{local *__ANON__ = "htc_$fname";\n} : '';
 
     no warnings 'uninitialized';
-    my $output = '$OUT .= ';
+    my $string_output = '$OUT .= ';
+    my $fh_output = 'print $OFH ';
+    my $output = $string_output;
     my $out_fh = $self->get_out_fh;
     if ($out_fh) {
-        $output = 'print $OFH ';
+        $output = $fh_output;
     }
+    my @outputs = ($output);
     my $header = <<"EOM";
 sub {
     use vars qw/ \$__ix__ \$__key__ \$__value__ \$__break__ \$__size__ /;
     use strict;
     no warnings;
 $anon
-    my (\$t, \$P, \$C, \$OFH) = \@_;
+    my (\$t, \$P, \$C, \$OFH, \$args) = \@_;
     my \$OUT = '';
 EOM
 
@@ -453,6 +488,7 @@ EOM
         lexicals       => \@lexicals,
     );
     my %use_vars;
+    my @wrapped;
     for my $token (@p) {
         @use_vars{ @lexicals } = () if @lexicals;
         my ($text, $line, $open_close, $tname, $attr, $f, $nlevel) = @$token;
@@ -600,7 +636,7 @@ EOM
                 expr        => $expr,
             );
             $code .= <<"EOM";
-${indent}local \$lexi_$var = $varstr;
+${indent}local \$HTML::Template::Compiled::_lexi_$var = $varstr;
 EOM
         }
         # --------- TMPL_LOOP TMPL_WHILE TMPL_EACH
@@ -652,7 +688,7 @@ EOM
             }
             my $global = '';
             my $lexi =
-              defined $lexical ? "${indent}local \$lexi_$lexical = \$\$C;\n" : "";
+              defined $lexical ? "${indent}local \$HTML::Template::Compiled::_lexi_$lexical = \$\$C;\n" : "";
             if ($self->get_global_vars) {
                 my $pop_global = _expr_method(
                     'pushGlobalstack',
@@ -832,7 +868,7 @@ EOM
         }
 
         # --------- TMPL_INCLUDE_VAR
-        elsif ($tname eq T_INCLUDE_VAR or $tname eq T_INCLUDE) {
+        elsif ($tname eq T_INCLUDE_VAR or $tname eq T_INCLUDE or $tname eq T_WRAPPER) {
             my $filename;
             my $varstr;
             my $path = $self->get_path();
@@ -840,6 +876,18 @@ EOM
             my $dynamic = $tname eq T_INCLUDE_VAR ? 1 : 0;
             my $fullpath = "''";
 
+            my $cwd;
+            unless ($self->get_scalar) {
+                $dir      = dirname($fname);
+                if ($self->get_search_path == 1) {
+                }
+                elsif ($self->get_search_path == 2) {
+                    $cwd = $dir;
+                }
+                else {
+                    $path = [ $dir ] ;
+                }
+            }
             if ($dynamic) {
                 # dynamic filename
                 my $dfilename = $attr->{NAME};
@@ -857,34 +905,15 @@ EOM
             else {
                 # static filename
                 $filename = $attr->{NAME};
+                $fullpath = $self->createFilename( [@$path], \$filename, $cwd );
                 if ($self->get_use_query) {
                     $info_stack->[-1]->{lc $filename}->{type} = $tname;
                 }
                 $varstr   = $self->quote_file($filename);
-                my $cwd;
-				unless ($self->get_scalar) {
-					$dir      = dirname($fname);
-					if ($self->get_search_path) {
-						if ( defined $dir and !grep { $dir eq $_ } @$path ) {
-                            $cwd = $dir;
-							# add the current directory to top of paths
-							# create new $path, don't alter original ref
-#							$path = [ $dir, @$path ] ;
-						}
-					}
-					else {
-							$path = [ $dir ] ;
-					}
-				}
                 # generate included template
                 {
                     D && $self->log("compile include $filename!!");
-                    #warn __PACKAGE__.':'.__LINE__.$".Data::Dumper->Dump([\$self->get_file], ['file']);
-                    #warn __PACKAGE__.':'.__LINE__.$".Data::Dumper->Dump([\$path], ['path']);
-                    #$fullpath = $self->createFilename( [@$path, \$self->get_file], $filename );
-                    $fullpath = $self->createFilename( [@$path], $filename, $cwd );
                     my $recursed = ++$HTML::Template::Compiled::COMPILE_STACK{$fullpath};
-                    #warn __PACKAGE__." fullpath $fullpath ($recursed)\n";
                     if ($recursed <= 1) {
                         my $cached_or_new;
                         $self->compile_early() and $cached_or_new
@@ -892,8 +921,6 @@ EOM
                                 #[@$path, \$self->get_file], $filename, '', $self->get_cache_dir
                               $path, $filename, '', $self->get_cache_dir
                           );
-                          #$fullpath = $cached_or_new->get_file;
-                        #$HTML::Template::Compiled::COMPILE_STACK{"@$path/$filename"} = $fullpath;
                         $self->get_includes()->{$fullpath}
                             = [$path, $filename, $cached_or_new];
                     }
@@ -907,47 +934,87 @@ EOM
               ? '['
               . join( ',', map { $self->quote_file($_) } @$path ) . ']'
               : 'undef';
+            $cwd = defined $cwd ? $self->quote_file($cwd) : 'undef';
             $cache = defined $cache ? $self->quote_file($cache) : 'undef';
             if ($dynamic) {
                 $code .= <<"EOM";
-${indent}\{
-${indent}  if (defined (my \$file = $varstr)) \{
-            my \$recursed = ++\$HTML::Template::Compiled::FILESTACK{$fullpath};
-            #warn "recursed \$recursed\\n";
-            \$HTML::Template::Compiled::FILESTACK{$fullpath} = 0, die "HTML::Template: recursive include of " . $fullpath . " \$recursed times (max \$HTML::Template::Compiled::MAX_RECURSE)"
-              if \$recursed > \$HTML::Template::Compiled::MAX_RECURSE;
-${indent}    my \$include = \$t->get_includes()->{$fullpath};
-${indent}    my \$new = \$include ? \$include->[2] : undef;
-#print STDERR "+++++++got new? \$new\\n";
-${indent}    if (!\$new || HTML::Template::Compiled::needs_new_check($cache||'',\$file)) {
-${indent}      \$new = \$t->new_from_object($path,\$file,$fullpath,$cache);
-${indent}    }
-#print STDERR "got new? \$new\\n";
-${indent}    \$new->set_globalstack(\$t->get_globalstack);
-${indent}    $output \$new->get_code()->(\$new,\$P,\$C@{[$out_fh ? ",\$OFH" : '']});
-            --\$HTML::Template::Compiled::FILESTACK{$fullpath} or delete \$HTML::Template::Compiled::FILESTACK{$fullpath};
-${indent}  \}
-${indent}\}
+# ---------- INCLUDE_VAR
+\{
+  if (defined (my \$file = $varstr)) \{
+    my \$fullpath = \$t->createFilename( $path, \\\$file, $cwd );
+    my \$recursed = ++\$HTML::Template::Compiled::FILESTACK{\$fullpath};
+    \$HTML::Template::Compiled::FILESTACK{\$fullpath} = 0, die "HTML::Template: recursive include of " . \$fullpath . " \$recursed times (max \$HTML::Template::Compiled::MAX_RECURSE)"
+      if \$recursed > \$HTML::Template::Compiled::MAX_RECURSE;
+    my \$include = \$t->get_includes()->{\$fullpath};
+    my \$new = \$include ? \$include->[2] : undef;
+    if (!\$new || HTML::Template::Compiled::needs_new_check($cache||'',\$file,\$t->get_expire_time)) \{
+      \$new = \$t->new_from_object($path,\$file,\$fullpath,$cache);
+    \}
+    \$new->set_globalstack(\$t->get_globalstack);
+    $output \$new->get_code()->(\$new,\$P,\$C@{[$out_fh ? ",\$OFH" : '']});
+    --\$HTML::Template::Compiled::FILESTACK{\$fullpath} or delete \$HTML::Template::Compiled::FILESTACK{\$fullpath};
+  \}
+\}
 EOM
             }
+            elsif ($tname eq T_WRAPPER) {
+                push @outputs, '$OUT' . (1 + scalar @outputs) . ' .= ';
+                $output = $outputs[-1];
+                my $wrapped = '';
+                $code .= <<"EOM";
+# ---------- WRAPPER
+\{
+  my \$OUT@{[ scalar @outputs ]};
+EOM
+                my $argument_fh = 'undef';
+                if ($out_fh) {
+                    $wrapped .= <<"EOM";
+my \$tmp_var = '';
+open my \$tmp_fh, '>', \\\$tmp_var;
+EOM
+                    $argument_fh = "\$tmp_fh";
+                }
+                $wrapped .= <<"EOM";
+  my \$_WRAPPED = \$OUT@{[ scalar @outputs ]};
+  my \$recursed = ++\$HTML::Template::Compiled::FILESTACK{$fullpath};
+  \$HTML::Template::Compiled::FILESTACK{$fullpath} = 0, die "HTML::Template: recursive include of " . $fullpath . " \$recursed times (max \$HTML::Template::Compiled::MAX_RECURSE)"
+  if \$recursed > \$HTML::Template::Compiled::MAX_RECURSE;
+    my \$include = \$t->get_includes()->{$fullpath};
+    my \$new = \$include ? \$include->[2] : undef;
+    if (!\$new) {
+      \$new = \$t->new_from_object($path,$varstr,$fullpath,$cache);
+    }
+    \$new->set_globalstack(\$t->get_globalstack);
+    $outputs[-2] \$new->get_code()->(\$new,\$P,\$C, $argument_fh, { wrapped => \$_WRAPPED });
+    --\$HTML::Template::Compiled::FILESTACK{$fullpath} or delete \$HTML::Template::Compiled::FILESTACK{$fullpath};
+  \$OUT@{[ scalar @outputs ]} = '';
+EOM
+                if ($out_fh) {
+                    $wrapped .= <<"EOM";
+$outputs[-2] \$tmp_var;
+EOM
+                }
+                $wrapped .= <<"EOM";
+\}
+EOM
+                push @wrapped, $wrapped;
+            }
             else {
                 $code .= <<"EOM";
-${indent}\{
-            my \$recursed = ++\$HTML::Template::Compiled::FILESTACK{$fullpath};
-            #warn "+recursed \$recursed $fullpath\\n";
-            \$HTML::Template::Compiled::FILESTACK{$fullpath} = 0, die "HTML::Template: recursive include of " . $fullpath . " \$recursed times (max \$HTML::Template::Compiled::MAX_RECURSE)"
-              if \$recursed > \$HTML::Template::Compiled::MAX_RECURSE;
-${indent}    my \$include = \$t->get_includes()->{$fullpath};
-${indent}    my \$new = \$include ? \$include->[2] : undef;
-#print STDERR "got new? \$new\\n";
-${indent}    if (!\$new) {
-${indent}      \$new = \$t->new_from_object($path,$varstr,$fullpath,$cache);
-${indent}    }
-#print STDERR "got new? \$new\\n";
-${indent}    \$new->set_globalstack(\$t->get_globalstack);
-${indent}    $output \$new->get_code()->(\$new,\$P,\$C@{[$out_fh ? ",\$OFH" : '']});
-            --\$HTML::Template::Compiled::FILESTACK{$fullpath} or delete \$HTML::Template::Compiled::FILESTACK{$fullpath};
-${indent}\}
+# ---------- INCLUDE
+\{
+  my \$recursed = ++\$HTML::Template::Compiled::FILESTACK{$fullpath};
+  \$HTML::Template::Compiled::FILESTACK{$fullpath} = 0, die "HTML::Template: recursive include of " . $fullpath . " \$recursed times (max \$HTML::Template::Compiled::MAX_RECURSE)"
+  if \$recursed > \$HTML::Template::Compiled::MAX_RECURSE;
+    my \$include = \$t->get_includes()->{$fullpath};
+    my \$new = \$include ? \$include->[2] : undef;
+    if (!\$new) {
+      \$new = \$t->new_from_object($path,$varstr,$fullpath,$cache);
+    }
+    \$new->set_globalstack(\$t->get_globalstack);
+    $output \$new->get_code()->(\$new,\$P,\$C@{[$out_fh ? ",\$OFH" : '']});
+    --\$HTML::Template::Compiled::FILESTACK{$fullpath} or delete \$HTML::Template::Compiled::FILESTACK{$fullpath};
+\}
 EOM
             }
         }
@@ -999,6 +1066,14 @@ ${indent}\$t->popGlobalstack;
 EOM
             }
         }
+        elsif ($tname eq T_WRAPPER) {
+            $code .= $wrapped[-1];
+            pop @wrapped;
+pop @outputs;
+$output = $outputs[-1];
+            $code .= <<"EOM";
+EOM
+        }
         else {
             # user defined
             #warn Data::Dumper->Dump([\$token], ['token']);
@@ -1018,14 +1093,14 @@ EOM
     }
     my @use_vars = grep length, keys %use_vars;
     if (@use_vars) {
-        $header .= qq#use vars qw/ @{[ map { '$lexi_'.$_ } @use_vars ]} /;\n#;
+#        $header .= qq#use vars qw/ @{[ map { '$_lexi_'.$_ } @use_vars ]} /;\n#;
     }
     #warn Data::Dumper->Dump([\$info], ['info']);
     $code .= qq#return \$OUT;\n#;
     $code = $header . $code . "\n} # end of sub\n";
 
     #$code .= "\n} # end of sub\n";
-    print STDERR "# ----- code \n$code\n# end code\n" if $self->get_debug;
+    print STDERR "# ----- code \n$code\n# end code\n" if $self->get_debug->{options} & HTML::Template::Compiled::DEBUG_COMPILED();
 
     # untaint code
     if ( $code =~ m/(\A.*\z)/ms ) {
